PREF = 100000.0  # Reference pressure for surface: Pa
RHOREF = 1.2  # Reference density for surface: kg / m+3
SCALE_HEIGHT = 7000  # Reference Scale-height: m

GAS_CONST_DRY_AIR = 287.058  # Gas constant for dry air: J / K / kg
SPEC_HEAT_DRY_AIR = 1004.64  # Spec. heat for dry air (const pressure): J / K / kg

EARTH_RADIUS = 6.37123e6  # Radius of Earth: m
EARTH_ROTA_RATE = 7.29211e-5  # Rotation rate of Earth: s-1
